"use strict";﻿

let fs = require("fs");
let path = require("path");
let readdirp = require("readdirp");
let configs = require("../conf/configs");
let errors = require("../errors");

let webRootList = [];
let webRootScanDone = false;

exports.checkAndAddWebRoot = (entry) =>
{
	if ((!entry.fullPath.endsWith(path.sep + "webroot" + path.sep)) &&
		(!entry.fullPath.endsWith(path.sep + "webroot"))) {
		return;
	}

	for (let p of webRootList) {
		if (p === entry.fullPath) {
			return;
		}
	}

	webRootList.push(entry.fullPath);
	configs.accessible.push(entry.fullPath);
	console.info("Added web root " + entry.fullPath);
};

exports.webRootScanDone = () =>
{
	webRootScanDone = true;
};

let fileCache = new Map();
let fileRealPathCache = new Map();

function _waitWebRootScan(callback)
{
	if (webRootScanDone) {
		return callback();
	}

	setTimeout(() => _waitWebRootScan(callback), 1000);
}

exports.getWebRoots = () =>
{
	return new Promise((resolve, reject) => {
		_waitWebRootScan(() => resolve(webRootScanDone));
	});
};

exports.getFileRealPath = (filePath) =>
{
	return new Promise((resolve, reject) => {
		_waitWebRootScan(() => {
			if (fileRealPathCache.has(filePath)) {
				return resolve(fileRealPathCache.get(filePath));
			}

			for (let r of webRootList) {
				let fullPath = path.normalize(path.join(r, filePath));

				if (!fullPath.startsWith(path.normalize(r))) {
					return reject(new errors.OutOfWebRoot("Requested path " + filePath + " escaped web root " + r));
				}

				let stat;

				try {
					stat = fs.statSync(fullPath);
				} catch (err) {
					continue;
				}

				if (stat.isFile) {
					fileRealPathCache.set(filePath, fullPath);
					return resolve(fullPath);
				}
			}

			reject(new Error("Cannot find " + filePath + " in any web root!"));
		});
	});
};

exports.getFile = (filePath, isExecutable, changedCallback) =>
{
	return new Promise((resolve, reject) => {
		_waitWebRootScan(() => {
			if (fileCache.has(filePath)) {
				return resolve(fileCache.get(filePath));
			}

			exports.getFileRealPath(filePath).then((fullPath) => {
				if (!isExecutable) {
					let data = fs.readFileSync(fullPath, "utf-8");
					fileCache.set(filePath, { data: data, realPath: fullPath });

					if (changedCallback) {
						try {
							fs.watch(fullPath, { presist: false, recursive: false }, (event, filename) => {
								console.info("Watched file " + filename + " has " + event);

								if (!filename) {
									return;
								}

								fileCache.delete(filePath);
								changedCallback(event, filename);
							});
						} catch (err) {
							console.error("Error occured when watching file " + filePath + " (" + fullPath + "): " + err.message);
							console.error(err.stack);
						}
					}

					return resolve({ data: data, realPath: fullPath });
				} else {
					let executor = require(fullPath);
					return resolve({ data: executor, realPath: fullPath });
				}
			}).catch((err) => reject(err));
		});
	});
};
