/**
 * Error types
 */

"use strict";

let util = require("util");

/**
 * Abstract class for all errors
 *
 * @param {string} msg - Error message
 * @param {function} constructor - Error object constructor
 */

let AbstractError = function (msg, constructor) {
	Error.captureStackTrace(this, constructor || this);
	this.message = msg || "Error";
};

util.inherits(AbstractError, Error);
AbstractError.prototype.name = "AbstractError";

/**
 * Error when a file is not found
 *
 * @param {string} msg - Error message
 */
let FileNotFoundError = function (msg) {
	FileNotFoundError.super_.call(this, msg, this.constructor);
};

util.inherits(FileNotFoundError, AbstractError);
FileNotFoundError.prototype.name = "FileNotFoundError";

/**
 * Error when a route cannot be routed to any controller
 *
 * @param {string} msg - Error message
 */

let NoRouteToControllerError = function (msg) {
	NoRouteToControllerError.super_.call(this, msg, this.constructor);
};

util.inherits(NoRouteToControllerError, AbstractError);
NoRouteToControllerError.prototype.name = "NoRouteToControllerError";

/**
 * Error when a route cannot be routed to any service
 *
 * @param {string} msg - Error message
 */

let NoRouteToServiceError = function (msg) {
	NoRouteToServiceError.super_.call(this, msg, this.constructor);
};

util.inherits(NoRouteToServiceError, AbstractError);
NoRouteToServiceError.prototype.name = "NoRouteToServiceError";

/**
 * Error when a controller doesn't have specified action
 *
 * @param {string} msg - Error message
 */
let NoSuchActionError = function (msg) {
	NoSuchActionError.super_.call(this, msg, this.constructor);
};

util.inherits(NoSuchActionError, AbstractError);
NoSuchActionError.prototype.name = "NoSuchActionError";

/**
 * Error when an action cannot be accessed with requested method
 *
 * @param {string} msg - Error message
 */
let MethodNotAllowedError = function (msg) {
	MethodNotAllowedError.super_.call(this, msg, this.constructor);
};

util.inherits(MethodNotAllowedError, AbstractError);
MethodNotAllowedError.prototype.name = "MethodNotAllowedError";

/**
 * Error when an action is not authencated
 *
 * @param {string} msg - Error message
 */
let ActionNoAuthError = function (msg) {
	ActionNoAuthError.super_.call(this, msg, this.constructor);
};

util.inherits(ActionNoAuthError, AbstractError);
ActionNoAuthError.prototype.name = "ActionNoAuthError";

/**
 * Error when a POST request length is too long
 *
 * @param {string} msg - Error message
 */
let RequestTooLongError = function (msg) {
	RequestTooLongError.super_.call(this, msg, this.constructor);
};

util.inherits(RequestTooLongError, AbstractError);
RequestTooLongError.prototype.name = "RequestTooLongError";

/**
 * Error when an XML file is not a valid view file
 *
 * @param {string} msg - Error message
 */
let InvalidViewFileError = function (msg) {
	InvalidViewFileError.super_.call(this, msg, this.constructor);
};

util.inherits(InvalidViewFileError, AbstractError);
InvalidViewFileError.prototype.name = "InvalidViewFileError";

/**
 * Error when a requested url points to a file which is out of any web roots
 *
 * @param {string} msg - Error message
 */
let OutOfWebRootError = function (msg) {
	OutOfWebRootError.super_.call(this, msg, this.constructor);
};

util.inherits(OutOfWebRootError, AbstractError);
OutOfWebRootError.prototype.name = "OutOfWebRootError";

/**
 * Error when a functor is invalid
 *
 * @param {string} msg - Error message
 */
let InvalidFunctorError = function (msg) {
	InvalidFunctorError.super_.call(this, msg, this.constructor);
};

util.inherits(InvalidFunctorError, AbstractError);
InvalidFunctorError.prototype.name = "InvalidFunctorError";

/**
 * Error when a filter rejected a request
 *
 * @param {string} msg - Error message
 */
let FilterRejectedError = function (msg) {
	FilterRejectedError.super_.call(this, msg, this.constructor);
};

util.inherits(FilterRejectedError, AbstractError);
FilterRejectedError.prototype.name = "FilterRejectedError";

/**
 * Error when security checker rejected a request
 *
 * @param {string} msg - Error message
 */
let InsecureRequestError = function (msg) {
	InsecureRequestError.super_.call(this, msg, this.constructor);
};

util.inherits(InsecureRequestError, AbstractError);
InsecureRequestError.prototype.name = "InsecureRequestError";

exports.FileNotFound = FileNotFoundError;
exports.NoRouteToController = NoRouteToControllerError;
exports.NoRouteToService = NoRouteToServiceError;
exports.NoSuchAction = NoSuchActionError;
exports.MethodNotAllowed = MethodNotAllowedError;
exports.ActionNoAuth = ActionNoAuthError;
exports.RequestTooLong = RequestTooLongError;
exports.InvalidViewFile = InvalidViewFileError;
exports.OutOfWebRoot = OutOfWebRootError;
exports.InvalidFunctor = InvalidFunctorError;
exports.FilterRejected = FilterRejectedError;
exports.InsecureRequest = InsecureRequestError;
