/**
 * HTTP Server for node.js
 */

"use strict";

let http = require("http");

let configs = require("../conf/configs");

let controllerLoader = require("../controller/loader");
let filterLoader = require("../filter/loader");
let serviceManager = require("../services/service-manager");
let router = require("./router");
let errors = require("../errors");
let readdirp = require("readdirp");
let path = require("path");
let domain = require("domain").create();
let viewEngine = require("../view/view-engine");
let webFiles = require("../caches/web-files");

let filterList = null;

function processRequest(request, response)
{
	let needResp = true;

	response.on("finish", function () {
		needResp = false;
	});

	router.route(request, response, controllerLoader.getControllerList(), filterLoader.getFilterList(),
		function (err, stream, mimeType, length, skipHeader) {
			if (err) {
				if (err instanceof errors.FileNotFound) {
					console.warn("FileNotFoundError: " + err.message);

					if (needResp) {
						response.writeHead(404);
						response.write(err.message);
						response.end();
					}

					return;
				}

				console.error("Error: " + err.name + ": " + err.message);
				console.error("Stacktrace:");
				console.error(err.stack);

				if (needResp) {
					response.writeHead(500);
					response.write(err.message);
					response.end();
				}

				return;
			}

			if (needResp) {
				if (!skipHeader) {
					response.writeHead(200, {
						"Content-Type": mimeType,
						"Content-Length": length
					});
				}

				if (stream) {
					stream.pipe(response);
				} else {
					response.end();
				}
			}
		});
}

let fileScanCallbacks = [];

exports.addFileScanCallback = (fileType, callback) =>
{
	fileScanCallbacks.push({
		typeRegEx: fileType,
		callback: callback
	});
};

/**
 * Start HTTP server
 *
 * @param {function} callback - Callback function with signature fn(configs: Object)
 */
exports.start = () =>
{
	return new Promise((resolve, reject) => {
		global.__node_bubble_authenticators__ = {};
		global.__node_bubble_view_models__ = {};

		readdirp({ root: configs.webRoot, entryType: "both" })
			.on("data", (entry) => {
				if (entry.fullPath.endsWith(".auth.js")) {
					let a = require(entry.fullPath);
					let name = a.getAuthenticatorName();

					if (!global.__node_bubble_authenticators__[name]) {
						global.__node_bubble_authenticators__[name] = a;
						console.info("Loaded authenticator at " + entry.path + " as name " + name);
					}
				} else if (entry.fullPath.endsWith(".controller.js")) {
					controllerLoader.loadController(entry.fullPath);
				} else if (entry.fullPath.endsWith(".filter.js")) {
					filterLoader.loadFilter(entry.fullPath);
				} else if (entry.fullPath.endsWith(".service.js")) {
					serviceManager.loadService(entry.fullPath);
				} else if ((entry.fullPath.endsWith(".tag.js")) || (entry.fullPath.endsWith(".tag.xml"))) {
					viewEngine.addTag(entry);
				} else if (entry.fullPath.endsWith(".view_model.js")) {
					let m = require(entry.fullPath);
					let name = m.getViewModelName();

					if (!global.__node_bubble_view_models__[name]) {
						global.__node_bubble_view_models__[name] = m;
						console.info("Loaded view model at " + entry.path + " as name " + name);
					}
				}

				if (entry.stat.isDirectory()) {
					webFiles.checkAndAddWebRoot(entry);
				}

				for (let c of fileScanCallbacks) {
					if (c.typeRegEx.test(entry.name)) {
						try {
							c.callback(entry, false);
						} catch (err) {
							console.warn("Error occured when scanning file with callback: " + err.message);
							console.warn(err.stack);
						}
					}
				}
			})
			.on("end", () => {
				viewEngine.tagScanDone();
				webFiles.webRootScanDone();
				console.info("Directory scan done.");

				for (let c of fileScanCallbacks) {
					c.callback(null, true);
				}

				resolve(configs);
			})
			.on("error", (err) => {
				console.error("Error: " + err);
				console.error(err.stack);
			});

		domain.on("error", (err) => {
			console.error("Uncaught error: " + err);
			console.error(err.stack);
			console.error("Please fix this and restart your application!");
		});

		domain.run(() => {
			let server = http.createServer(processRequest);
			server.listen(configs.port);

			console.info("Server " + configs.host + " in web root " + path.resolve(configs.webRoot) + " started at port " +
				configs.port);
		});
	});
};
