/**
 * Router for HTTP server
 */

"use strict";

let urlParser = require("url");
let fs = require("fs");
let path = require("path");
let mime = require("mime");

let configs = require("../conf/configs");

let controllerProcessor = require("../controller/processor");
let serviceManager = require("../services/service-manager");
let errors = require("../errors");
let accessLogger = require("../utils/logger").accessLogger;
let requestUtils = require("../utils/request");
let viewEngine = require("../view/view-engine");
let Cookies = require("cookies");
let streamUtils = require("../utils/stream");
let webFiles = require("../caches/web-files");
require("../utils/annotation");

function resolveFilePath(url)
{
	if (configs.enableWebRootChar) {
		let l = url.lastIndexOf("/" + configs.webRootChar + "/");
		if (l >= 0) {
			url = url.substring(l + 2);
		}
	}
	
	let realPath;
	let mapFound = false;
	let filePath;
	let allow = false;

	for (let key in configs.fileUrlMap) {
		if (url.indexOf(key) === 0) {
			url = url.replace(key, configs.fileUrlMap[key]);
			realPath = path.resolve(path.join(path.sep, configs.fileUrlMap[key]));
			mapFound = true;
			break;
		}
	}

	return new Promise((resolve, reject) => {
		if (mapFound) {
			filePath = path.resolve(path.join(path.sep, url));
			allow = (filePath.indexOf(realPath) === 0);

			return resolve({ filePath: filePath, allow: allow });
		}

		webFiles.getFileRealPath(url).then((p) => {
			resolve({ filePath: p, allow: true });
		}).catch((err) => {
			console.error(err);
			filePath = path.resolve(path.join(configs.webRoot, url));

			if (filePath.indexOf(path.resolve(configs.webRoot)) < 0) {
				console.warn("Path " + filePath + " is out of web root " + configs.webRoot + ", denied!");
				return resolve({ filePath: null, allow: false });
			}

			for (let i = 0; i < configs.accessible.length; i++) {
				let allowedPath = path.resolve(path.join(configs.webRoot, configs.accessible[i]));
				if ((filePath.indexOf(allowedPath) === 0) ||
					((path.resolve(configs.accessible[i]) === path.normalize(configs.accessible[i])) && (filePath.indexOf(configs.accessible[i]) === 0))) {
					allow = true;
				}
			}

			return resolve({ filePath: filePath, allow: allow });
		});
	});
}

/**
 * Route URL to file
 *
 * @param {object} urlData - The URL data to route
 * @param {object} request - The request object
 * @param {object} response - The response object
 * @param {function} callback - Callback function with signature fn(err: Error, stream: ReadStream,
 *								mimeType: string, length: integer)
 */
function routeToFile(urlData, request, response, callback)
{
	resolveFilePath(urlData.pathname).then((resolved) => {
		if (!resolved.allow) {
			console.warn("Path " + resolved.filePath + " is not found or not allowed to access!");
			return callback(new errors.FileNotFound("Requested url " + urlData.pathname + " not found!"));
		}

		fs.stat(resolved.filePath, (err, stats) => {
			if (err) {
				return callback(err);
			}

			if (request.headers["if-modified-since"]) {
				let modifyCheckDate = Date.parse(request.headers["if-modified-since"]);

				if (modifyCheckDate >= stats.mtime) {
					response.writeHead(304);
					response.end();
					callback(null, null, null, 0, true);
					return;
				}
			}

			response.setHeader("Last-Modified", stats.mtime.toUTCString());

			/*console.info("Requested url " + urlData.pathname + " mapped to file " + filePath
					+ ", length " + stats.size);*/

			callback(null, fs.createReadStream(resolved.filePath), mime.lookup(resolved.filePath), stats.size);
		});
	});
}

/**
 * DEPRECATED - Route URL to controller
 *
 * @param {object} request - The request
 * @param {object} response - The response
 * @param {string} urlData - The URL data to route
 * @param {object} controllerList - List of controllers
 * @param {function} callback - Callback function with signature fn(err: Error, stream: ReadStream,
 *								mimeType: string, length: integer)
 */
function routeToController(request, response, urlData, controllerList, callback)
{
	let c = null;

	for (let urlPattern in controllerList) {
		if (new RegExp(urlPattern, "i").test(urlData.pathname)) {
			c = controllerList[urlPattern];

			if (controllerProcessor.canProcess(urlData, c.controller)) {
				//console.info("Request " + urlData.pathname + " routed to controller " + controller);
				break;
			} else {
				c = null;
			}
		}
	}

	if (c == null) {
		callback(new errors.NoRouteToController("Request " + urlData.pathname + " cannot be routed to any controller!"));

		return;
	}

	controllerProcessor.process(request, response, urlData, c, callback);
}

function routeToService(request, response, urlData, callback)
{
	let serviceList = serviceManager.getLoadedServices();
	let service = null;

	for (let pattern in serviceList.keys()) {
		if (new RegExp(pattern, "i").test(urlData.pathname)) {
			service = serviceList.get(pattern);
			break;
		}
	}

	if ((service == null) || (service._service.handoff)) {
		return callback(new errors.NoRouteToService("Request " + urlData.pathname + " cannot be routed to any service!"));
	}

	serviceManager.doService(service, request, response, urlData, callback);
}

function routeToViewEngine(request, response, urlData, callback)
{
	requestUtils.getRequestData(request, (err, data, files) => {
		if (err) {
			callback(err);
			return;
		}

		viewEngine.executeView(urlData.pathname, {
			request: request,
			response: response,
			params: data,
			files: files,
			cookies: new Cookies(request, response),
			configs: configs
		})
		.then((data) => {
			let length = Buffer.byteLength(data, "utf8");
			let stream = streamUtils.dataToStream(data);
			callback(null, stream, "text/html", length);
		})
		.catch((err) => callback(err));
	});
}

function routeToFunctor(request, response, urlData, callback)
{
	requestUtils.getRequestData(request, (err, data, files) => {
		if (err) {
			callback(err);
			return;
		}

		webFiles.getFile(urlData.pathname, true).then((fileData) => {
			let context = {
				request: request,
				response: response,
				params: data,
				files: files,
				cookies: new Cookies(request, response),
				configs: configs
			};

			let functor = fileData.data;

			if (typeof functor !== "function") {
				return callback(new errors.InvalidFunctor("Functor " + urlData.pathname + " is not a function!"));
			}

			let actionInfo = functor.getAnnotationData("action");

			if (actionInfo.allow) {
				if (!(actionInfo.allow instanceof Array)) {
					actionInfo.allow = [ actionInfo.allow ];
				}

				let found = false;

				for (let a of actionInfo.allow) {
					if (a.toLowerCase() === request.method.toLowerCase()) {
						found = true;
						break;
					}
				}

				if (!found) {
					return callback(new errors.MethodNotAllowed("Action " + urlData.pathname + " doesn't like " + request.method));
				}
			}

			if (!actionInfo.insecure) {
				if ((!requestUtils.checkRequestParameters(data)) || (!requestUtils.checkRequestParameters(request.headers.cookies))) {
					return callback(new errors.InsecureRequest("Insecure data detected in request!"));
				}
			}

			let authPromise = new Promise((resolve, reject) => {
				if (actionInfo.auth) {
					if (!(actionInfo.auth instanceof Array)) {
						actionInfo.auth = [ actionInfo.auth ];
					}

					let authContext = {
						actionName: urlData.pathname,
						params: urlData.query,
						cookies: context.cookies,
						request: request,
						response: response
					};

					let authPassedCount = 0;

					let authCallback = (err, auth) => {
						if (!auth) {
							return reject(new errors.ActionNoAuth("Action " + urlData.pathname + " not authenticated!"));
						} else {
							authPassedCount++;

							if (err) {
								return reject(err);
							} else {
								if (authPassedCount === actionInfo.auth.length) {
									resolve();
								}
							}
						}
					};

					for (let a of actionInfo.auth) {
						let ss = a.split(".");
						let modName = ss[0];
						let funcName = ss[1];
						let authFunc;

						try {
							authFunc = global.__node_bubble_authenticators__[modName][funcName];
						} catch (err) {
							console.warn("Authenticator " + a + " is not a function or not found!");
							return authCallback(err, false);
						}

						if (typeof authFunc !== "function") {
							console.warn("Authenticator " + a + " is not a function or not found!");
							return authCallback(null, false);
						}

						authFunc(authContext).then((auth) => authCallback(null, auth)).catch((err) => authCallback(err, false));
					}
				} else {
					resolve();
				}
			});

			authPromise.then(() => {
				let returnType;
				let isJson = false;

				if (functor.hasAnnotation("returnJson")) {
					returnType = "application/json";
					isJson = true;
				} else {
					returnType = functor.getAnnotationData("return").type;
				}

				try {
					functor(context)
					.then((d) => {
						if (isJson) {
							d = JSON.stringify(d);
						}

						let length = Buffer.byteLength(d, "utf8");
						let stream = streamUtils.dataToStream(d);
						callback(null, stream, returnType, length);
					})
					.catch((err) => callback(err));
				} catch (err) {
					callback(err);
				}
			}).catch((err) => callback(err));
		}).catch((err) => callback(err));
	});
}

function executeFilters(filterList, action, request, response)
{
	return new Promise((resolve, reject) => {
		let executor = function*() {
			if ((filterList) && (filterList.length > 0)) {
				for (let i = 0; i < filterList.length; i++) {
					let result = yield filterList[i][action](request, response);
					if (result === false) {
						console.info("Request cut by filter " + filterList[i]);
						return reject(new errors.FilterRejected("Request cut by filter!"));
					}
				}
			}
		};

		let iter = executor();

		let worker = (d) => {
			let result = iter.next(d);

			if (!result.done) {
				result.value.then((r) => worker(d)).catch((err) => reject(err));
			} else {
				resolve();
			}
		};

		worker();
	});
}

/**
 * Route request to processors
 *
 * @param {object} request - The request to route
 * @param {object} response - The response
 * @param {object} controllerList - List of controllers
 * @param {Array} filterList - List of filters
 * @param {function} callback - Callback function with signature fn(err: Error, stream: ReadStream,
 *								mimeType: string, length: integer)
 */
exports.route = (request, response, controllerList, filterList, callback) =>
{
	accessLogger.info(requestUtils.getRequestIP(request) + " " + request.method + " " + request.url + " (" + request.headers["user-agent"] + ")");

	executeFilters(filterList, "beforeAction", request, response).then(() => {
		let urlData = urlParser.parse(request.url, true);
		urlData.pathname = urlData.pathname.replace("/" + configs.prefix + "/", "/");
		
		if (urlData.pathname === "/") {
			urlData.pathname = path.join(urlData.pathname, configs.defaultUrl);
		}

		let afterAction = (err, stream, mimeType, length) => {
			if (err) {
				return callback(err);
			}

			executeFilters(filterList, "afterAction", request, response).then(() => {
				callback(err, stream, mimeType, length);
			}).catch((err) => callback(err));
		};

		routeToService(request, response, urlData, (serr, stream, mimeType, length) => {
			if (serr) {
				if (serr instanceof errors.NoRouteToService) {
					routeToController(request, response, urlData, controllerList, (err, stream, mimeType, length) => {
							if (err) {
								if (err instanceof errors.NoRouteToController) {
									if (urlData.pathname.endsWith(".xml")) {
										routeToViewEngine(request, response, urlData, (verr, stream, mimeType, length) => {
											if (verr instanceof errors.InvalidViewFile) {
												routeToFile(urlData, request, response, afterAction);
												return;
											}

											afterAction(verr, stream, mimeType, length);
										});
									} else if (urlData.pathname.endsWith(".functor.js")) {
										routeToFunctor(request, response, urlData, afterAction);
										return;
									} else {
										routeToFile(urlData, request, response, afterAction);
										return;
									}
								} else {
									return callback(err);
								}
							} else {
								afterAction(null, stream, mimeType, length);
							}
						});
				} else {
					return callback(serr);
				}
			} else {
				afterAction(null, stream, mimeType, length);
			}
		});
	}).catch((err) => callback(err));
};
