/**
 * Request processing utils
 */

"use strict";

let configs = require("../conf/configs");
let errors = require("../errors");
let querystring = require("querystring");
let urlParser = require("url");
let multiparty = require("multiparty");
let xss = require("xss");

/**
 * Get data from request
 *
 * @param {object} request - The request object
 * @param {function} callback - Callback with signature fn(err: Error, data: Object, files: Object)
 */
exports.getRequestData = (request, callback) =>
{
	if (request.method === "GET") {
		callback(null, urlParser.parse(request.url, true).query);
	} else if (request.method === "POST") {
		let buf = "";
		let isMultipart = false;

		if ((request.headers["content-type"]) && (request.headers["content-type"].indexOf("multipart/form-data") >= 0)) {
			isMultipart = true;
		}

		if (!isMultipart) {
			request.setEncoding("utf8");
			request.on("data", (data) => {
				buf += data;

				if (buf.length > configs.maxPostSize) {
					request.connection.destroy();
					callback(new errors.RequestTooLong("Request exceeded max POST length limit!"));
					return;
				}
			})
			.on("end", () => {
				callback(null, querystring.parse(buf));
			});
		} else {
			let form = new multiparty.Form({ maxFilesSize: configs.maxPostSize });
			form.parse(request, (err, fields, files) => {
				if (err) {
					callback(err);
					return;
				}

				let params = urlParser.parse(request.url, true).query;
				for (let key in params) {
					if (!params.hasOwnProperty(key)) {
						continue;
					}

					if (fields[key]) {
						if (!(fields[key] instanceof Array)) {
							let a = [ fields[key] ];
							fields[key] = a;
						}

						fields[key].push(params[key]);
					} else {
						fields[key] = params[key];
					}
				}

				callback(null, fields, files);
			});
		}
	}
};

exports.getRequestIP = (req) =>
{
	try {
		return req.headers['x-forwarded-for'] || req.connection.remoteAddress || req.socket.remoteAddress ||
			req.connection.socket.remoteAddress;
	} catch (err) {
		return "";
	}
};

exports.checkRequestParameters = (params) =>
{
	if (configs.insecure) {
		return true;
	}

	if (typeof params === "object") {
		for (let key of Object.keys(params)) {
			if (key != xss(key)) {
				console.warn("Insecure request parameter name '" + key + "' detected!");
				return false;
			}

			if (params[key] != xss(params[key])) {
				console.warn("Insecure request parameter value '" + params[key] + "' of key '" + key + "' detected!");
				return false;
			}
		}
	} else if (typeof params === "string") {
		return (params === xss(params));
	}

	return true;
};
