"use strict";﻿

let xmldoc = require("xmldoc");
let fs = require("fs");
let path = require("path");
let readdirp = require("readdirp");
let renderer = require("./renderer-async");
let errors = require("../errors");
let webFiles = require("../caches/web-files");
let requestUtils = require("../utils/request");

let tagList = new Map();
let xmlTagList = new Map();
let tagScanDone = false;

exports.addTag = (entry) =>
{
	if (!entry.stat.isFile) {
		return;
	}

	let tagName = entry.name.substring(0, entry.name.indexOf(".tag."));

	if (entry.name.endsWith(".tag.js")) {
		let Tag = require(entry.fullPath);
		let tagInstance = new Tag();

		if (tagList.has(tagName)) {
			return;
		}

		tagList.set(tagName, { tag: tagInstance, dir: path.dirname(entry.fullPath) });

		console.info("Added tag " + entry.fullPath);
	} else if (entry.name.endsWith(".tag.xml")) {
		if (xmlTagList.has(tagName)) {
			return;
		}

		try {
			let xmlTag = fs.readFileSync(entry.fullPath, "utf-8");
			if ((!xmlTag) || (!xmlTag.startsWith("<!DOCTYPE bubble-tag>"))) {
				console.warn("XML tag file " + entry.fullPath + " is not a valid XML tag!");
				return;
			}

			xmlTagList.set(tagName, { tag: xmlTag, dir: path.dirname(entry.fullPath) });

			console.info("Added tag " + entry.fullPath);
		} catch (err) {
			console.error("Error occured when adding tag " + entry.fullPath + ": " + err.message);
			console.error(err.stack);
		}
	}
};

exports.tagScanDone = () =>
{
	tagScanDone = true;
};

function _waitTagScan(callback)
{
	if (tagScanDone) {
		return callback();
	}

	setTimeout(() => _waitTagScan(callback), 1000);
}

function _processXmlTag(tagData, attrs, body, deferredList, viewDir)
{
	return new Promise((resolve, reject) => {
		let data = tagData;

		if (typeof attrs === "object") {
			for (let a of Object.keys(attrs)) {
				let regExp = new RegExp("<\\!--\\[# " + a + " \\]-->", "g");
				data = data.replace(regExp, attrs[a]);
			}
		}

		body = body.replace(/\$/g, "$$$$");
		data = data.replace(new RegExp("<\\!--\\[# body \\]-->", "g"), body).replace(/<\!--\[#(.*?)\]-->/g, "undefined");

		let document;

		try {
			document = new xmldoc.XmlDocument(data);
		} catch (err) {
			console.error("Error occured when processing XML document: " + err.message);
			console.error(err.stack);
			console.error("Current document: ");
			console.error(data);
			return reject(err);
		}

		_expandNode(document, deferredList, viewDir).then((d) => resolve(d)).catch((err) => reject(err));
	});
}

function _renderTag(name, attrs, body, deferredList, viewDir)
{
	return new Promise((resolve, reject) => {
		let tagInfo = tagList.get(name);
		if (!tagInfo) {
			tagInfo = xmlTagList.get(name);
			if (!tagInfo) {
				console.error("Unknown tag " + name + " found!");
				resolve({ content: "", deferredList: deferredList });
			}

			_processXmlTag(tagInfo.tag, attrs, body, deferredList, viewDir).then((data) => resolve({ content: data.content, deferredList: data.deferredList })).catch((err) => reject(err));
		} else {
			let renderedData;

			try {
				renderedData = tagInfo.tag._render(attrs, body, deferredList, viewDir, tagInfo.dir);
			} catch (err) {
				console.error("Error occured when rendering tag " + name + ": " + err.message);
				console.error(err.stack);
				return reject(err);
			}

			resolve({ content: renderedData.content, deferredList: renderedData.deferredList });
		}
	});
}

function _expandNode(xmlNode, deferredList, viewDir)
{
	return new Promise((resolve, reject) => {
		let expandChildOnly = false;
		let isPart = false;

		if (xmlNode.name === "raw-html") {
			return resolve({ content: xmlNode.val, deferredList: deferredList });
		} else if (xmlNode.name === "bview:part") {
			expandChildOnly = true;
			isPart = true;
		} else if (xmlNode.name.startsWith("bview:part-")) {
			return resolve({ content: "", deferredList: deferredList });
		} else if ((!xmlNode.children) || (xmlNode.children.length <= 0)) {
			return _renderTag(xmlNode.name, xmlNode.attr, xmlNode.val, deferredList, viewDir)
				.then((data) => resolve({ content: data.content, deferredList: data.deferredList }))
				.catch((err) => reject(err));
		}

		let realContent = "";
		
		let _iterateChild = function* () {
			for (let child of xmlNode.children) {
				yield _expandNode(child, deferredList, viewDir);
			}
		};

		let iter = _iterateChild();

		let worker = () => {
			let result = iter.next();

			if (!result.done) {
				result.value.then((d) => {
					realContent += d.content;
					worker();
				}).catch((err) => reject(err));
			} else {
				if (!expandChildOnly) {
					if (xmlNode.name === "root") {
						resolve({ content: realContent, deferredList: deferredList });
					} else {
						_renderTag(xmlNode.name, xmlNode.attr, realContent, deferredList, viewDir)
							.then((data) => resolve({ content: data.content, deferredList: data.deferredList }))
							.catch((err) => reject(err));
					}
				} else {
					if (isPart) {
						realContent = "<bview:part-" + xmlNode.attr.name + ">" + realContent + "</bview:part-" + xmlNode.attr.name + ">\n";
					}

					resolve({ content: realContent, deferredList: deferredList });
				}
			}
		};

		worker();
	});
}

let viewCache = new Map();
let clientBridges = {};

function _compile(viewPath, content)
{
	return new Promise((resolve, reject) => {
		let document;

		try {
			document = new xmldoc.XmlDocument(content);
		} catch (err) {
			console.error("Error occured when compiling view " + viewPath + ": " + err.message);
			console.error(err.stack);
			reject(err);
		}

		let deferredList = [];

		_waitTagScan(() => {
			_expandNode(document, deferredList, path.normalize(path.dirname(viewPath))).then((data) => {
				let o = renderer.render(data.content, clientBridges);
				resolve({ executor: o.cache, deferredList: data.deferredList });
			}).catch((err) => reject(err));
		});
	});
};

exports.registerClientBridge = (name, object) =>
{
	clientBridges[name] = object;
};

exports.getClientBridges = () =>
{
	return clientBridges;
};

exports.executeView = (viewPath, requestContext) =>
{
	let viewExecutor = null;
	let viewDeferredList = null;

	let viewCompiledPromise = new Promise((resolve, reject) => {
		if (!viewCache.has(viewPath)) {
			webFiles.getFile(viewPath, false, () => viewCache.delete(viewPath))
			.then((viewData) => {
				if (!viewData.data.startsWith("<!DOCTYPE bubble-view>")) {
					return reject(new errors.InvalidViewFile(viewData.realPath + " is not a valid view file!"));
				}

				if (viewData.data.indexOf("<!-- INSECURE -->") < 0) {
					if ((!requestUtils.checkRequestParameters(requestContext.params)) || (!requestUtils.checkRequestParameters(requestContext.request.headers.cookies))) {
						return reject(new errors.InsecureRequest("Insecure data detected in request!"));
					}
				}

				_compile(viewData.realPath, viewData.data).then((viewCacheInfo) => {
					viewExecutor = viewCacheInfo.executor;
					viewDeferredList = viewCacheInfo.deferredList;
					viewCache.set(viewPath, viewCacheInfo);
					resolve();
				}).catch((err) => reject(err));
			})
			.catch((err) => reject(err));
		} else {
			let viewCacheInfo = viewCache.get(viewPath);
			viewExecutor = viewCacheInfo.executor;
			viewDeferredList = viewCacheInfo.deferredList;

			resolve();
		}
	});

	let viewExecutedPromise = new Promise((resolve, reject) => {
		viewCompiledPromise.then(() => {
			let executor = viewExecutor(null, requestContext, viewDeferredList);

			let worker = (data) => {
				let result;

				try {
					result = executor.next(data);
				} catch (err) {
					reject(err);
				}

				if (result.value instanceof Promise) {
					result.value.then((d) => worker(d)).catch((err) => {
						console.error("Error occured when executing view: " + err.message);
						console.error(err.stack);
						reject(err);
					});
				} else if (result.done) {
					resolve(result.value);
				} else {
					worker(result.value);
				}
			};

			worker();
		}).catch((err) => reject(err));
	});

	return viewExecutedPromise;
};
